import bencode from "bencode";
import { promises as fs } from "fs";
import parseTorrent from "parse-torrent";
import { dirname, resolve } from "path";
import { inspect } from "util";
import xmlrpc from "xmlrpc";
import { InjectionResult } from "../constants.js";
import { CrossSeedError } from "../errors.js";
import { Label, logger } from "../logger.js";
import { getRuntimeConfig } from "../runtimeConfig.js";
import { wait } from "../utils.js";
async function createLibTorrentResumeTree(meta, dataDir) {
    async function getFileResumeData(file) {
        const filePath = resolve(dataDir, file.path);
        const fileStat = await fs
            .lstat(filePath)
            .catch(() => ({ isFile: () => false }));
        if (!fileStat.isFile() || fileStat.size !== file.length) {
            logger.debug({
                label: Label.RTORRENT,
                message: `File ${filePath} either doesn't exist or is the wrong size.`,
            });
            return {
                completed: 0,
                mtime: 0,
                priority: 0,
            };
        }
        return {
            completed: Math.ceil(file.length / meta.pieceLength),
            mtime: Math.trunc(fileStat.mtimeMs / 1000),
            priority: 0,
        };
    }
    return {
        bitfield: Math.ceil(meta.length / meta.pieceLength),
        files: await Promise.all(meta.files.map(getFileResumeData)),
    };
}
async function saveWithLibTorrentResume(meta, savePath, dataDir) {
    const rawMeta = bencode.decode(parseTorrent.toTorrentFile(meta));
    rawMeta.libtorrent_resume = await createLibTorrentResumeTree(meta, dataDir);
    await fs.writeFile(savePath, bencode.encode(rawMeta));
}
export default class RTorrent {
    constructor() {
        const { rtorrentRpcUrl } = getRuntimeConfig();
        try {
            const { origin, username, password, protocol, pathname } = new URL(rtorrentRpcUrl);
            const clientCreator = protocol === "https:"
                ? xmlrpc.createSecureClient
                : xmlrpc.createClient;
            const shouldUseAuth = Boolean(username && password);
            this.client = clientCreator({
                url: origin + pathname,
                basic_auth: shouldUseAuth
                    ? {
                        user: decodeURIComponent(username),
                        pass: decodeURIComponent(password),
                    }
                    : undefined,
            });
        }
        catch (e) {
            throw new CrossSeedError("rTorrent url must be percent-encoded");
        }
    }
    async methodCallP(method, args) {
        logger.verbose({
            label: Label.RTORRENT,
            message: `Calling method ${method} with params ${inspect(args, {
                depth: null,
                compact: true,
            })}`,
        });
        return new Promise((resolve, reject) => {
            this.client.methodCall(method, args, (err, data) => {
                if (err)
                    return reject(err);
                return resolve(data);
            });
        });
    }
    async checkForInfoHashInClient(infoHash) {
        const downloadList = await this.methodCallP("download_list", []);
        return downloadList.includes(infoHash.toUpperCase());
    }
    async getTorrentConfiguration(searchee) {
        if (!searchee.infoHash) {
            throw new CrossSeedError("rTorrent direct injection not implemented for data-based searchees");
        }
        const infoHash = searchee.infoHash.toUpperCase();
        const result = await this.methodCallP("system.multicall", [
            [
                {
                    methodName: "d.is_multi_file",
                    params: [infoHash],
                },
                {
                    methodName: "d.directory",
                    params: [infoHash],
                },
                {
                    methodName: "d.complete",
                    params: [infoHash],
                },
            ],
        ]);
        // temp diag for #154
        try {
            const [[isMultiFileStr], [dir], [isCompleteStr]] = result;
            return {
                dataDir: Number(isMultiFileStr) ? dirname(dir) : dir,
                isComplete: Boolean(Number(isCompleteStr)),
            };
        }
        catch (e) {
            logger.error(e);
            logger.debug("Failure caused by server response below:");
            logger.debug(inspect(result));
        }
    }
    async validateConfig() {
        const { rtorrentRpcUrl } = getRuntimeConfig();
        // no validation to do
        if (!rtorrentRpcUrl)
            return;
        try {
            await this.methodCallP("download_list", []);
        }
        catch (e) {
            logger.debug(e);
            throw new CrossSeedError(`Failed to reach rTorrent at ${rtorrentRpcUrl}`);
        }
    }
    async inject(meta, searchee, nonceOptions) {
        const { outputDir: runtimeConfigOutputDir } = getRuntimeConfig();
        const { outputDir = runtimeConfigOutputDir } = nonceOptions;
        if (await this.checkForInfoHashInClient(meta.infoHash)) {
            return InjectionResult.ALREADY_EXISTS;
        }
        const { dataDir, isComplete } = await this.getTorrentConfiguration(searchee);
        if (!isComplete)
            return InjectionResult.TORRENT_NOT_COMPLETE;
        const torrentFilePath = resolve(outputDir, `${meta.name}.tmp.${Date.now()}.torrent`);
        await saveWithLibTorrentResume(meta, torrentFilePath, dataDir);
        for (let i = 0; i < 5; i++) {
            try {
                await this.methodCallP("load.start", [
                    "",
                    torrentFilePath,
                    `d.directory.set="${dataDir}"`,
                    `d.custom1.set="cross-seed"`,
                    `d.custom.set=addtime,${Math.round(Date.now() / 1000)}`,
                ]);
                break;
            }
            catch (e) {
                await wait(1000 * Math.pow(2, i));
            }
        }
        for (let i = 0; i < 5; i++) {
            if (await this.checkForInfoHashInClient(meta.infoHash)) {
                setTimeout(() => fs.unlink(torrentFilePath), 1000);
                return InjectionResult.SUCCESS;
            }
            await wait(100 * Math.pow(2, i));
        }
        setTimeout(() => fs.unlink(torrentFilePath), 1000);
        return InjectionResult.FAILURE;
    }
}
//# sourceMappingURL=RTorrent.js.map