import ms from "ms";
import fetch from "node-fetch";
import xml2js from "xml2js";
import { EP_REGEX, SEASON_REGEX } from "./constants.js";
import { db } from "./db.js";
import { CrossSeedError } from "./errors.js";
import { getEnabledIndexers, IndexerStatus, updateIndexerStatus, } from "./indexers.js";
import { Label, logger } from "./logger.js";
import { getRuntimeConfig } from "./runtimeConfig.js";
import { cleanseSeparators, getTag, MediaType, nMsAgo, reformatTitleForSearching, stripExtension, } from "./utils.js";
function sanitizeUrl(url) {
    url = new URL(url);
    return url.origin + url.pathname;
}
function getApikey(url) {
    return new URL(url).searchParams.get("apikey");
}
function parseTorznabResults(xml) {
    var _a, _b, _c;
    const items = (_c = (_b = (_a = xml === null || xml === void 0 ? void 0 : xml.rss) === null || _a === void 0 ? void 0 : _a.channel) === null || _b === void 0 ? void 0 : _b[0]) === null || _c === void 0 ? void 0 : _c.item;
    if (!items || !Array.isArray(items)) {
        return [];
    }
    return items.map((item) => {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j;
        return ({
            guid: item.guid[0],
            name: item.title[0],
            tracker: (_j = (_f = (_c = (_b = (_a = item === null || item === void 0 ? void 0 : item.prowlarrindexer) === null || _a === void 0 ? void 0 : _a[0]) === null || _b === void 0 ? void 0 : _b._) !== null && _c !== void 0 ? _c : (_e = (_d = item === null || item === void 0 ? void 0 : item.jackettindexer) === null || _d === void 0 ? void 0 : _d[0]) === null || _e === void 0 ? void 0 : _e._) !== null && _f !== void 0 ? _f : (_h = (_g = item === null || item === void 0 ? void 0 : item.indexer) === null || _g === void 0 ? void 0 : _g[0]) === null || _h === void 0 ? void 0 : _h._) !== null && _j !== void 0 ? _j : "Unknown tracker",
            link: item.link[0],
            size: Number(item.size[0]),
            pubDate: new Date(item.pubDate[0]).getTime(),
        });
    });
}
function parseTorznabCaps(xml) {
    var _a, _b;
    const capsSection = (_b = (_a = xml === null || xml === void 0 ? void 0 : xml.caps) === null || _a === void 0 ? void 0 : _a.searching) === null || _b === void 0 ? void 0 : _b[0];
    const isAvailable = (searchTechnique) => { var _a, _b; return ((_b = (_a = searchTechnique === null || searchTechnique === void 0 ? void 0 : searchTechnique[0]) === null || _a === void 0 ? void 0 : _a.$) === null || _b === void 0 ? void 0 : _b.available) === "yes"; };
    return {
        search: Boolean(isAvailable(capsSection === null || capsSection === void 0 ? void 0 : capsSection.search)),
        tvSearch: Boolean(isAvailable(capsSection === null || capsSection === void 0 ? void 0 : capsSection["tv-search"])),
        movieSearch: Boolean(isAvailable(capsSection === null || capsSection === void 0 ? void 0 : capsSection["movie-search"])),
    };
}
function createTorznabSearchQuery(name, caps) {
    const nameWithoutExtension = stripExtension(name);
    const extractNumber = (str) => parseInt(str.match(/\d+/)[0]);
    const mediaType = getTag(nameWithoutExtension);
    if (mediaType === MediaType.EPISODE && caps.tvSearch) {
        const match = nameWithoutExtension.match(EP_REGEX);
        return {
            t: "tvsearch",
            q: cleanseSeparators(match.groups.title),
            season: extractNumber(match.groups.season),
            ep: extractNumber(match.groups.episode),
        };
    }
    else if (mediaType === MediaType.SEASON && caps.tvSearch) {
        const match = nameWithoutExtension.match(SEASON_REGEX);
        return {
            t: "tvsearch",
            q: cleanseSeparators(match.groups.title),
            season: extractNumber(match.groups.season),
        };
    }
    else {
        return {
            t: "search",
            q: reformatTitleForSearching(nameWithoutExtension),
        };
    }
}
export async function queryRssFeeds() {
    const candidatesByUrl = await makeRequests("", await getEnabledIndexers(), () => ({ t: "search", q: "" }));
    return candidatesByUrl.flatMap((e) => e.candidates);
}
export async function searchTorznab(name) {
    const { excludeRecentSearch, excludeOlder } = getRuntimeConfig();
    const enabledIndexers = await getEnabledIndexers();
    // search history for name across all indexers
    const timestampDataSql = await db("searchee")
        .join("timestamp", "searchee.id", "timestamp.searchee_id")
        .join("indexer", "timestamp.indexer_id", "indexer.id")
        .whereIn("indexer.id", enabledIndexers.map((i) => i.id))
        .andWhere({ name })
        .select({
        indexerId: "indexer.id",
        firstSearched: "timestamp.first_searched",
        lastSearched: "timestamp.last_searched",
    });
    const indexersToUse = enabledIndexers.filter((indexer) => {
        const entry = timestampDataSql.find((entry) => entry.indexerId === indexer.id);
        return (!entry ||
            ((!excludeOlder || entry.firstSearched > nMsAgo(excludeOlder)) &&
                (!excludeRecentSearch ||
                    entry.lastSearched < nMsAgo(excludeRecentSearch))));
    });
    const timestampCallout = " (filtered by timestamps)";
    logger.info({
        label: Label.TORZNAB,
        message: `Searching ${indexersToUse.length} indexers for ${name}${indexersToUse.length < enabledIndexers.length
            ? timestampCallout
            : ""}`,
    });
    return makeRequests(name, indexersToUse, (indexer) => createTorznabSearchQuery(name, {
        search: indexer.searchCap,
        tvSearch: indexer.tvSearchCap,
        movieSearch: indexer.movieSearchCap,
    }));
}
export async function syncWithDb() {
    const { torznab } = getRuntimeConfig();
    const dbIndexers = await db("indexer")
        .where({ active: true })
        .select({
        id: "id",
        url: "url",
        apikey: "apikey",
        active: "active",
        status: "status",
        retryAfter: "retry_after",
        searchCap: "search_cap",
        tvSearchCap: "tv_search_cap",
        movieSearchCap: "movie_search_cap",
    });
    const inConfigButNotInDb = torznab.filter((configIndexer) => !dbIndexers.some((dbIndexer) => dbIndexer.url === sanitizeUrl(configIndexer)));
    const inDbButNotInConfig = dbIndexers.filter((dbIndexer) => !torznab.some((configIndexer) => sanitizeUrl(configIndexer) === dbIndexer.url));
    const apikeyUpdates = dbIndexers.reduce((acc, dbIndexer) => {
        const configIndexer = torznab.find((configIndexer) => sanitizeUrl(configIndexer) === dbIndexer.url);
        if (configIndexer &&
            dbIndexer.apikey !== getApikey(configIndexer)) {
            acc.push({
                id: dbIndexer.id,
                apikey: getApikey(configIndexer),
            });
        }
        return acc;
    }, []);
    if (inDbButNotInConfig.length > 0) {
        await db("indexer")
            .whereIn("url", inDbButNotInConfig.map((indexer) => indexer.url))
            .update({ active: false });
    }
    if (inConfigButNotInDb.length > 0) {
        await db("indexer")
            .insert(inConfigButNotInDb.map((url) => ({
            url: sanitizeUrl(url),
            apikey: getApikey(url),
            active: true,
        })))
            .onConflict("url")
            .merge(["active", "apikey"]);
    }
    await db.transaction(async (trx) => {
        for (const apikeyUpdate of apikeyUpdates) {
            await trx("indexer")
                .where({ id: apikeyUpdate.id })
                .update({ apikey: apikeyUpdate.apikey });
        }
    });
}
function assembleUrl(urlStr, apikey, params) {
    const url = new URL(urlStr);
    const searchParams = new URLSearchParams();
    searchParams.set("apikey", apikey);
    for (const [key, value] of Object.entries(params)) {
        if (value != null)
            searchParams.set(key, value);
    }
    url.search = searchParams.toString();
    return url.toString();
}
function fetchCaps(indexer) {
    return fetch(assembleUrl(indexer.url, indexer.apikey, { t: "caps" }))
        .then((r) => r.text())
        .then(xml2js.parseStringPromise)
        .then(parseTorznabCaps);
}
function collateOutcomes(correlators, outcomes) {
    return outcomes.reduce(({ rejected, fulfilled }, cur, idx) => {
        if (cur.status === "rejected") {
            rejected.push([correlators[idx], cur.reason]);
        }
        else {
            fulfilled.push([correlators[idx], cur.value]);
        }
        return { rejected, fulfilled };
    }, { rejected: [], fulfilled: [] });
}
async function updateCaps(indexers) {
    const outcomes = await Promise.allSettled(indexers.map((indexer) => fetchCaps(indexer)));
    const { fulfilled, rejected } = collateOutcomes(indexers.map((i) => i.id), outcomes);
    for (const [indexerId, reason] of rejected) {
        logger.warn(`Failed to reach ${indexers.find((i) => i.id === indexerId).url}`);
        logger.debug(reason);
    }
    for (const [indexerId, caps] of fulfilled) {
        await db("indexer").where({ id: indexerId }).update({
            search_cap: caps.search,
            tv_search_cap: caps.tvSearch,
            movie_search_cap: caps.movieSearch,
        });
    }
}
export async function validateTorznabUrls() {
    const { torznab } = getRuntimeConfig();
    if (!torznab)
        return;
    const urls = torznab.map((str) => new URL(str));
    for (const url of urls) {
        if (!url.pathname.endsWith("/api")) {
            throw new CrossSeedError(`Torznab url ${url} must have a path ending in /api`);
        }
        if (!url.searchParams.has("apikey")) {
            throw new CrossSeedError(`Torznab url ${url} does not specify an apikey`);
        }
    }
    await syncWithDb();
    const enabledIndexersWithoutCaps = await db("indexer")
        .where({
        active: true,
        search_cap: null,
        tv_search_cap: null,
        movie_search_cap: null,
    })
        .select({ id: "id", url: "url", apikey: "apikey" });
    await updateCaps(enabledIndexersWithoutCaps);
    const indexersWithoutSearch = await db("indexer")
        .where({ search_cap: false, active: true })
        .select({ id: "id", url: "url" });
    for (const indexer of indexersWithoutSearch) {
        logger.warn(`Ignoring indexer that doesn't support searching: ${indexer.url}`);
    }
    const indexersWithSearch = await getEnabledIndexers();
    if (indexersWithSearch.length === 0) {
        throw new CrossSeedError("no working indexers available");
    }
}
async function makeRequests(name, indexers, getQuery) {
    const searchUrls = indexers.map((indexer) => assembleUrl(indexer.url, indexer.apikey, getQuery(indexer)));
    searchUrls.forEach((message) => void logger.verbose({ label: Label.TORZNAB, message }));
    const abortControllers = searchUrls.map(() => new AbortController());
    setTimeout(() => {
        for (const abortController of abortControllers) {
            abortController.abort();
        }
    }, 30000).unref();
    const outcomes = await Promise.allSettled(searchUrls.map((url, i) => fetch(url, {
        headers: { "User-Agent": "cross-seed" },
        signal: abortControllers[i].signal,
    })
        .then((response) => {
        if (!response.ok) {
            if (response.status === 429) {
                updateIndexerStatus(IndexerStatus.RATE_LIMITED, Date.now() + ms("1 hour"), [indexers[i].id]);
            }
            else {
                updateIndexerStatus(IndexerStatus.UNKNOWN_ERROR, Date.now() + ms("1 hour"), [indexers[i].id]);
            }
            throw new Error(`request failed with code: ${response.status}`);
        }
        return response;
    })
        .then((r) => r.text())
        .then(xml2js.parseStringPromise)
        .then(parseTorznabResults)));
    const { rejected, fulfilled } = collateOutcomes(indexers.map((indexer) => indexer.id), outcomes);
    for (const [indexerId, reason] of rejected) {
        logger.warn(`Failed to reach ${indexers.find((i) => i.id === indexerId).url}`);
        logger.debug(reason);
    }
    return fulfilled.map(([indexerId, results]) => ({
        indexerId,
        candidates: results,
    }));
}
//# sourceMappingURL=torznab.js.map